import argparse
import json
import os
import sys

from hydromodel.calibrate.calibrate_ga import calibrate_by_ga

sys.path.append("../..")
from hydromodel.calibrate.calibrate_sceua import calibrate_by_sceua
from hydromodel.utils import hydro_utils


def main(args):
    data = hydro_utils.unserialize_numpy(
        os.path.join(args.data_dir, "basins_lump_p_pe_q.npy")
    )
    algo = args.algorithm
    warmup = args.warmup_length
    model = args.model_name
    algo_param = args.algorithm_param
    if algo == "SCE_UA":
        calibrate_by_sceua(
            data[:, :, 0:2],
            data[:, :, -1:],
            warmup_length=warmup,
            model=model,
            **algo_param
        )
    elif algo == "GA":
        calibrate_by_ga(
            data[:, :, 0:2],
            data[:, :, -1:],
            warmup_length=warmup,
            model=model,
            **algo_param
        )
    else:
        raise NotImplementedError(
            "We don't provide this calibrate method! Choose from 'SCE_UA' or 'GA'!"
        )


# python calibrate_xaj.py --data_dir "D:\\code\\hydro-model-xaj\\hydromodel\\example" --warmup_length 60
if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Calibrate XAJ model by SCE-UA.")
    parser.add_argument(
        "--data_dir",
        dest="data_dir",
        help="the data directory for XAJ model",
        default="../example",
        type=str,
    )
    parser.add_argument(
        "--warmup_length",
        dest="warmup_length",
        help="the length of warmup period for XAJ model",
        default=30,
        type=int,
    )
    parser.add_argument(
        "--model_name",
        dest="model_name",
        help="different implementations for XAJ: 'xaj' or 'xaj_mz'",
        default="xaj",
        type=str,
    )
    parser.add_argument(
        "--algorithm",
        dest="algorithm",
        help="calibrate algorithm: SCE_UA (default) or GA",
        default="SCE_UA",
        type=str,
    )
    parser.add_argument(
        "--algorithm_param",
        dest="algorithm_param",
        help="parameters set for calibrate algorithm",
        default={
            "random_seed": 2000,
            "rep": 5000,
            "ngs": 7,
            "kstop": 3,
            "peps": 0.1,
            "pcento": 0.1,
        },
        type=json.loads,
    )
    the_args = parser.parse_args()
    main(the_args)



import math
from typing import Optional, Tuple

import numpy as np
from numba import jit

from hydromodel.models.xaj import uh_conv


@jit
def calculate_precip_store(s, precip_net, x1):
    """Calculates the amount of rainfall which enters the storage reservoir."""
    n = x1 * (1.0 - (s / x1) ** 2) * np.tanh(precip_net / x1)
    d = 1.0 + (s / x1) * np.tanh(precip_net / x1)
    return n / d


@jit
def calculate_evap_store(s, evap_net, x1):
    """Determines the evaporation loss from the production store"""
    n = s * (2.0 - s / x1) * np.tanh(evap_net / x1)
    d = 1.0 + (1.0 - s / x1) * np.tanh(evap_net / x1)
    return n / d


@jit
def calculate_perc(current_store, x1):
    """Determines how much water percolates out of the production store to streamflow"""
    return current_store * (
        1.0 - (1.0 + (4.0 / 9.0 * current_store / x1) ** 4) ** -0.25
    )


def production(
    p_and_e: np.array, x1: np.array, s_level: Optional[np.array] = None
) -> Tuple[np.array, np.array]:
    """
    an one-step calculation for production store in GR4J
    the dimension of the cell: [batch, feature]
    Parameters
    ----------
    p_and_e
        P is pe[:, 0] and E is pe[:, 1]; similar with the "input" in the RNNCell
    x1:
        Storage reservoir parameter;
    s_level
        s_level means S in the GR4J Model; similar with the "hx" in the RNNCell
        Initial value of storage in the storage reservoir.
    Returns
    -------
    tuple
        contains the Pr and updated S
    """
    # Calculate net precipitation and evapotranspiration
    precip_difference = p_and_e[:, 0] - p_and_e[:, 1]
    precip_net = np.maximum(precip_difference, 0.0)
    evap_net = np.maximum(-precip_difference, 0.0)

    if s_level is None:
        s_level = 0.6 * x1

    # s_level should not be larger than x1
    s_level = np.clip(s_level, a_min=np.full(s_level.shape, 0.0), a_max=x1)

    # Calculate the fraction of net precipitation that is stored
    precip_store = calculate_precip_store(s_level, precip_net, x1)

    # Calculate the amount of evaporation from storage
    evap_store = calculate_evap_store(s_level, evap_net, x1)

    # Update the storage by adding effective precipitation and
    # removing evaporation
    s_update = s_level - evap_store + precip_store
    # s_level should not be larger than self.x1
    s_update = np.clip(s_update, a_min=np.full(s_update.shape, 0.0), a_max=x1)

    # Update the storage again to reflect percolation out of the store
    perc = calculate_perc(s_update, x1)
    s_update = s_update - perc
    # perc is always lower than S because of the calculation itself, so we don't need clamp here anymore.

    # The precip. for routing is the sum of the rainfall which
    # did not make it to storage and the percolation from the store
    current_runoff = perc + (precip_net - precip_store)
    return current_runoff, s_update


@jit
def s_curves1(t, x4):
    """
    Unit hydrograph ordinates for UH1 derived from S-curves.
    """

    if t <= 0:
        return 0
    elif t < x4:
        return (t / x4) ** 2.5
    else:  # t >= x4
        return 1


@jit
def s_curves2(t, x4):
    """
    Unit hydrograph ordinates for UH2 derived from S-curves.
    """

    if t <= 0:
        return 0
    elif t < x4:
        return 0.5 * (t / x4) ** 2.5
    elif t < 2 * x4:
        return 1 - 0.5 * (2 - t / x4) ** 2.5
    else:  # t >= x4
        return 1


def uh_gr4j(x4):
    """
    Generate the convolution kernel for the convolution operation in routing module of GR4J

    Parameters
    ----------
    x4
        the dim of x4 is [batch]
    Returns
    -------
    list
        UH1s and UH2s for all basins
    """
    uh1_ordinates = []
    uh2_ordinates = []
    for i in range(len(x4)):
        n_uh1 = int(math.ceil(x4[i]))
        n_uh2 = int(math.ceil(2.0 * x4[i]))
        uh1_ordinate = np.zeros(n_uh1)
        uh2_ordinate = np.zeros(n_uh2)
        for t in range(1, n_uh1 + 1):
            uh1_ordinate[t - 1] = s_curves1(t, x4[i]) - s_curves1(t - 1, x4[i])

        for t in range(1, n_uh2 + 1):
            uh2_ordinate[t - 1] = s_curves2(t, x4[i]) - s_curves2(t - 1, x4[i])
        uh1_ordinates.append(uh1_ordinate)
        uh2_ordinates.append(uh2_ordinate)

    return uh1_ordinates, uh2_ordinates


def routing(q9: np.array, q1: np.array, x2, x3, r_level: Optional[np.array] = None):
    """
    the GR4J routing-module unit cell for time-sequence loop
    Parameters
    ----------
    q9
    q1
    x2
        Catchment water exchange parameter
    x3
        Routing reservoir parameters
    r_level
        Beginning value of storage in the routing reservoir.
    Returns
    -------
    """
    if r_level is None:
        r_level = 0.7 * x3
    # r_level should not be larger than self.x3
    r_level = np.clip(r_level, a_min=np.full(r_level.shape, 0.0), a_max=x3)
    groundwater_ex = x2 * (r_level / x3) ** 3.5
    r_updated = np.maximum(np.full(r_level.shape, 0.0), r_level + q9 + groundwater_ex)

    qr = r_updated * (1.0 - (1.0 + (r_updated / x3) ** 4) ** -0.25)
    r_updated = r_updated - qr

    qd = np.maximum(np.full(groundwater_ex.shape, 0.0), q1 + groundwater_ex)
    q = qr + qd
    return q, r_updated


def gr4j(p_and_e, parameters, warmup_length: int, return_state=False):
    """
    run GR4J model

    Parameters
    ----------
    p_and_e

    parameters
        2-dim variable -- [basin, parameter]:
        the parameters are x1, x2, x3 and x4
    warmup_length
        length of warmup period
    return_state
        if True, return state values, mainly for warmup periods

    Returns
    -------
    Union[np.array, tuple]
        streamflow or (streamflow, states)
    """
    x1_scale = [100.0, 1200.0]
    x2_sacle = [-5.0, 3.0]
    x3_scale = [20.0, 300.0]
    x4_scale = [1.1, 2.9]
    x1 = x1_scale[0] + parameters[:, 0] * (x1_scale[1] - x1_scale[0])
    x2 = x2_sacle[0] + parameters[:, 1] * (x2_sacle[1] - x2_sacle[0])
    x3 = x3_scale[0] + parameters[:, 2] * (x3_scale[1] - x3_scale[0])
    x4 = x4_scale[0] + parameters[:, 3] * (x4_scale[1] - x4_scale[0])

    if warmup_length > 0:
        # set no_grad for warmup periods
        p_and_e_warmup = p_and_e[0:warmup_length, :, :]
        _, s0, r0 = gr4j(p_and_e_warmup, parameters, warmup_length=0, return_state=True)
    else:
        s0 = 0.5 * x1
        r0 = 0.5 * x3
    inputs = p_and_e[warmup_length:, :, :]
    streamflow_ = np.full(inputs.shape[:2], 0.0)
    prs = np.full(inputs.shape[:2], 0.0)
    for i in range(inputs.shape[0]):
        if i == 0:
            pr, s = production(inputs[i, :, :], x1, s0)
        else:
            pr, s = production(inputs[i, :, :], x1, s)
        prs[i, :] = pr
    prs_x = np.expand_dims(prs, axis=2)
    conv_q9, conv_q1 = uh_gr4j(x4)
    q9 = np.full([inputs.shape[0], inputs.shape[1], 1], 0.0)
    q1 = np.full([inputs.shape[0], inputs.shape[1], 1], 0.0)
    for j in range(inputs.shape[1]):
        q9[:, j : j + 1, :] = uh_conv(
            prs_x[:, j : j + 1, :], conv_q9[j].reshape(-1, 1, 1)
        )
        q1[:, j : j + 1, :] = uh_conv(
            prs_x[:, j : j + 1, :], conv_q1[j].reshape(-1, 1, 1)
        )
    for i in range(inputs.shape[0]):
        if i == 0:
            q, r = routing(q9[i, :, 0], q1[i, :, 0], x2, x3, r0)
        else:
            q, r = routing(q9[i, :, 0], q1[i, :, 0], x2, x3, r)
        streamflow_[i, :] = q
    streamflow = np.expand_dims(streamflow_, axis=2)
    if return_state:
        return streamflow, s, r
    return streamflow



"""TODO: river routing process"""
import numpy as np


def network_route(runoffs, route_params):
    """
    河网汇流计算，新安江模型里一般采用线性水库或滞后演算法。这里使用滞后演算法

    Parameters
    ----------
    runoffs
        坡面汇流计算结果，数组
    route_params
        模型参数

    Returns
    -------
    np.array
        汇流计算结果——流量过程线
    """
    # 取整后面才能计算
    t = int(route_params["L"])
    cr = route_params["CR"]
    # 初始化
    qr = runoffs
    q = np.zeros(runoffs.size)
    if t <= 0:
        for i in range(runoffs.size):
            if i == 0:
                q[0] = (1 - cr) * qr[0]
            else:
                q[i] = cr * q[i - 1] + (1 - cr) * qr[i]
    else:
        for i in range(runoffs.size):
            if i == 0:
                q[0] = 0
            elif i < t:
                q[i] = cr * q[i - 1]
            else:
                q[i] = cr * q[i - 1] + (1 - cr) * qr[i - t]
    return q


def river_route(config, route_params, qf):
    """
    河道汇流计算，新安江模型一般采用马斯京根法

    Parameters
    ----------
    config
        计算设置条件
    route_params
        模型参数
    qf
        河网汇流计算结果，数组

    Returns
    -------
    np.array
        汇流计算结果——流量过程线
    """
    ke = route_params["KE"]
    xe = route_params["XE"]
    time_interval = config["time_interval/h"]

    q = np.zeros(qf.size)

    c0 = (0.5 * time_interval - ke * xe) / (0.5 * time_interval + ke - ke * xe)
    c1 = (0.5 * time_interval + ke * xe) / (0.5 * time_interval + ke - ke * xe)
    c2 = 1 - c0 - c1

    q[0] = qf[0]
    if c0 >= 0 and c2 >= 0:
        for i in range(1, q.size):
            q[i] = c0 * qf[i] + c1 * qf[i - 1] + c2 * q[i - 1]
    else:
        # 当马斯京根不适用时，暂未处理
        q = qf
    return q


# -------------------------------------------Unfinished auxiliary function---------------------------------------------
def uh_recognise(runoffs, flood_data):
    """
    时段单位线的识别，先以最小二乘法为主。

    Parameters
    ----------
    runoffs
        各场次洪水对应的各时段净雨，np.array组成的list
    flood_data
        多组出口断面流量过程线，np.array组成的list

    Returns
    -------
    np.array
        时段单位线，每个数据的单位为m^3/s
    """
    # 最小二乘法计算针对每场次洪水得到一条单位线，多场次洪水，按照书上的意思，可以取平均。如果曲线之间差异较大，需要进行分类，目前先求平均。
    uh_s = []
    uh_sum = np.array([])
    for i in range(len(runoffs)):
        ht = []
        # q表示实际径流
        q = flood_data[i]
        l = len(flood_data[i])
        m = len(runoffs[i])
        n = l - m + 1
        for j in range(n):
            # numpy默认为行向量
            h_row = np.zeros(l)
            for k in range(j, j + m):
                h_row[k] = runoffs[i][k - j]
            ht.append(h_row)
        h = np.transpose(ht)
        ht_h = np.dot(ht, h)
        ht_q = np.dot(ht, q)
        # 求得一条单位线
        uh_temp = np.linalg.solve(ht_h, ht_q)
        # 每场次洪水均有一条单位线
        uh_s.append(uh_temp)
        # 求和，因为单位线长度可能不一致，所以需要进行补0对齐
        if i == 0:
            uh_sum = uh_temp
        else:
            # 当维度不同的向量要对齐时，需在不足处补0
            length_zero = max(uh_sum.size, uh_temp.size) - min(
                uh_sum.size, uh_temp.size
            )
            zeros_need = np.zeros(length_zero)
            if uh_sum.size > uh_temp.size:
                arr_new = np.hstack([uh_temp, zeros_need])
                uh_sum = uh_sum + arr_new
            else:
                arr_new = np.hstack([uh_sum, zeros_need])
                uh_sum = uh_temp + arr_new
    # 广播运算
    uh = uh_sum / len(runoffs)
    return uh


def split_flow(floods_data, auto=None):
    """
    分割流量过程，有时候，会手动进行分割，所以可能不需要调用次函数

    Parameters
    ----------
    floods_data
        多组观测的出口断面流量过程线，np.array组成的list
    auto
        是程序自动计算退水曲线还是已经手动计算好了

    Returns
    -------
    list[np.array]
        去掉前期洪水尚未退完的部分水量和非本次降雨补给的流量过程线
    """
    # TODO: 求退水曲线
    # 首先取出各个洪水过程退水段
    floods = []
    for flood_data in floods_data:
        temp = flood_data.loc[:, "flood_quant"]
        floods.append(np.array(temp))
    # 然后取出退水段各点，组成向量

    # 最小二乘法计算

    # 利用计算好的退水曲线分割观测的流量过程线

    if auto is None:
        return floods
    return floods


def divide_source(floods, auto=None):
    """
    从观测的场次洪水径流中分割地表地下径流

    Parameters
    ----------
    floods
        多组观测的已经去掉前期洪水尚未退完的部分水量和非本次降雨补给的出口断面流量过程线，np.array组成的list
    auto
        是程序自动计算分割流量还是已经手动计算好了

    Returns
    -------
    list[np.array]
        rs_s,rg_s: 地表径流流量过程线和地下径流过程线
    """
    # TODO: 分割径流
    if auto is None:
        return floods
    return floods


def iuh_recognise(
    runoffs, flood_data, linear_reservoir=None, linear_canal=None, isochrone=None
):
    """
    瞬时单位线的识别，目前以计算Nash单位线为主

    Parameters
    ----------
    runoffs
        多组各时段净雨（地表），矩阵表示
    flood_data
        多组出口断面流量过程线，矩阵表示
    linear_reservoir

    linear_canal
        线性渠个数
    isochrone
        等流时线个数

    Returns
    -------
    tuple[float, float]
        n,k:nash单位线两参数
    """
    # TODO: 这个函数没想好怎么写，能很好地把串并联结构和三个基础模型搭配在一起，目前以n个线性水库串联为主

    return



import spotpy
from matplotlib import pyplot as plt


def show_calibrate_result(spot_setup, result_file_name, flow_unit="mm day-1"):
    """
    Plot one year result to see the effect of optimized parameters

    Parameters
    ----------
    spot_setup
        Spotpy's setup class instance
    result_file_name
        the result file saved after optimizing
    flow_unit
        unit of streamflow

    Returns
    -------
    None
    """
    # Load the results gained with the sceua sampler, stored in SCEUA_xaj.csv
    results = spotpy.analyser.load_csv_results(result_file_name)
    # Plot how the objective function was minimized during sampling
    fig = plt.figure(1, figsize=(9, 6))
    plt.plot(results["like1"])
    plt.ylabel("RMSE")
    plt.xlabel("Iteration")
    # Plot the best model run
    # Find the run_id with the minimal objective function value
    bestindex, bestobjf = spotpy.analyser.get_minlikeindex(results)

    # Select best model run
    best_model_run = results[bestindex]

    # Filter results for simulation results
    fields = [word for word in best_model_run.dtype.names if word.startswith("sim")]
    best_simulation = list(best_model_run[fields])

    fig = plt.figure(figsize=(9, 6))
    ax = plt.subplot(1, 1, 1)
    # TODO: now we just plot one year's data
    ax.plot(
        best_simulation[365:730],
        color="black",
        linestyle="solid",
        label="Best objf.=" + str(bestobjf),
    )
    ax.plot(
        spot_setup.evaluation()[365:730], "r.", markersize=3, label="Observation data"
    )
    plt.xlabel("Number of Observation Points")
    plt.ylabel("Discharge [" + flow_unit + "]")
    plt.legend(loc="upper right")



import os

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pytest
import spotpy

from hydromodel.calibrate.calibrate_sceua import calibrate_by_sceua, SpotSetup
from hydromodel.calibrate.calibrate_ga import calibrate_by_ga
import definitions
from hydromodel.visual.pyspot_plots import show_calibrate_result
from hydromodel.models.xaj import xaj, uh_gamma, uh_conv


@pytest.fixture()
def basin_area():
    # the area of basin 01013500, unit km2
    # basin_area = 2252.7
    return 1.783


@pytest.fixture()
def warmup_length():
    return 30


@pytest.fixture()
def the_data():
    root_dir = definitions.ROOT_DIR
    # test_data = pd.read_csv(os.path.join(root_dir, "hydromodel", "example", '01013500_lump_p_pe_q.txt'))
    return pd.read_csv(
        os.path.join(root_dir, "hydromodel", "example", "hymod_input.csv"), sep=";"
    )


@pytest.fixture()
def p_and_e(the_data):
    # p_and_e_df = test_data[['prcp(mm/day)', 'petfao56(mm/day)']]
    # three dims: sequence (time), batch (basin), feature (variable)
    # p_and_e = np.expand_dims(p_and_e_df.values, axis=1)
    p_and_e_df = the_data[["rainfall[mm]", "TURC [mm d-1]"]]
    return np.expand_dims(p_and_e_df.values, axis=1)


@pytest.fixture()
def qobs(basin_area, the_data):
    # 1 ft3 = 0.02831685 m3
    ft3tom3 = 2.831685e-2
    # 1 km2 = 10^6 m2
    km2tom2 = 1e6
    # 1 m = 1000 mm
    mtomm = 1000
    # 1 day = 24 * 3600 s
    daytos = 24 * 3600
    # qobs_ = np.expand_dims(test_data[['streamflow(ft3/s)']].values, axis=1)
    # trans ft3/s to mm/day
    # return qobs_ * ft3tom3 / (basin_area * km2tom2) * mtomm * daytos

    qobs_ = np.expand_dims(the_data[["Discharge[ls-1]"]].values, axis=1)
    # trans l/s to mm/day
    return qobs_ * 1e-3 / (basin_area * km2tom2) * mtomm * daytos


@pytest.fixture()
def params():
    # all parameters are in range [0,1]
    return np.tile([0.5], (1, 15))


def test_uh_gamma():
    # repeat for 20 periods and add one dim as feature: time_seq=20, batch=10, feature=1
    routa = np.tile(2.5, (20, 10, 1))
    routb = np.tile(3.5, (20, 10, 1))
    uh = uh_gamma(routa, routb, len_uh=15)
    np.testing.assert_almost_equal(
        uh[:, 0, :],
        np.array(
            [
                [0.0069],
                [0.0314],
                [0.0553],
                [0.0738],
                [0.0860],
                [0.0923],
                [0.0939],
                [0.0919],
                [0.0875],
                [0.0814],
                [0.0744],
                [0.0670],
                [0.0597],
                [0.0525],
                [0.0459],
            ]
        ),
        decimal=3,
    )


def test_uh():
    uh_from_gamma = np.tile(1, (5, 3, 1))
    rf = np.arange(30).reshape(10, 3, 1) / 100
    qs = uh_conv(rf, uh_from_gamma)
    np.testing.assert_almost_equal(
        np.array(
            [
                [0.0000, 0.0100, 0.0200],
                [0.0300, 0.0500, 0.0700],
                [0.0900, 0.1200, 0.1500],
                [0.1800, 0.2200, 0.2600],
                [0.3000, 0.3500, 0.4000],
                [0.4500, 0.5000, 0.5500],
                [0.6000, 0.6500, 0.7000],
                [0.7500, 0.8000, 0.8500],
                [0.9000, 0.9500, 1.0000],
                [1.0500, 1.1000, 1.1500],
            ]
        ),
        qs[:, :, 0],
        decimal=3,
    )


def test_xaj(p_and_e, params, warmup_length):
    qsim = xaj(p_and_e, params, warmup_length=warmup_length)
    np.testing.assert_array_equal(qsim.shape[0], p_and_e.shape[0] - warmup_length)


def test_xaj_mz(p_and_e, params, warmup_length):
    qsim = xaj(p_and_e, params, warmup_length=warmup_length, route_method="MZ")
    np.testing.assert_array_equal(qsim.shape[0], p_and_e.shape[0] - warmup_length)


def test_calibrate_xaj_sceua(p_and_e, qobs, warmup_length):
    calibrate_by_sceua(
        p_and_e,
        qobs,
        warmup_length,
        random_seed=2000,
        rep=5000,
        ngs=7,
        kstop=3,
        peps=0.1,
        pcento=0.1,
    )


def test_show_calibrate_sceua_result(p_and_e, qobs, warmup_length):
    spot_setup = SpotSetup(
        p_and_e, qobs, warmup_length, obj_func=spotpy.objectivefunctions.rmse
    )
    show_calibrate_result(spot_setup, "SCEUA_xaj")
    plt.show()


def test_calibrate_xaj_ga(p_and_e, qobs, warmup_length):
    calibrate_by_ga(
        p_and_e,
        qobs,
        warmup_length,
        run_counts=5,
        pop_num=50,
        cross_prob=0.5,
        mut_prob=0.5,
    )



import numpy as np
import scipy.stats

from hydromodel.utils.hydro_utils import hydro_logger


def KGE(xs, xo) -> float:
    """
    Kling Gupta Efficiency (Gupta et al., 2009, http://dx.doi.org/10.1016/j.jhydrol.2009.08.003)

    Parameters
    ----------
    xs
        simulated
    xo
        observed

    Returns
    -------
    float
        KGE: Kling Gupta Efficiency
    """
    r = np.corrcoef(xo, xs)[0, 1]
    alpha = np.std(xs) / np.std(xo)
    beta = np.mean(xs) / np.mean(xo)
    kge = 1 - np.sqrt((r - 1) ** 2 + (alpha - 1) ** 2 + (beta - 1) ** 2)
    return kge


def NSE(xs, xo) -> float:
    """
    Calculate NSE

    Parameters
    ----------
    xs
        simulated
    xo
        observed

    Returns
    -------
    float
        NSE: Nash-Sutcliffe model efficiency coefficient
    """
    x_mean = xo.mean()
    SST = np.sum((xo - x_mean) ** 2)
    SSRes = np.sum((xo - xs) ** 2)
    nse = 1 - SSRes / SST
    return nse


def statNse(target, pred):
    """
    Calculate NSE for two dimensional array

    Parameters
    ----------
    target
        observation
    pred
        prediction

    Returns
    -------
    np.array
        one NSE for one element in the 1-dim result
    """
    ngrid, nt = pred.shape
    NSe = np.full(ngrid, np.nan)
    for k in range(0, ngrid):
        x = pred[k, :]
        y = target[k, :]
        ind = np.where(np.logical_and(~np.isnan(x), ~np.isnan(y)))[0]
        if ind.shape[0] > 0:
            xx = x[ind]
            yy = y[ind]
            NSe[k] = NSE(xx, yy)
    return NSe


def statRmse(target, pred, axis=0):
    """
    Calculate RMSE for multi-dim arrays

    Parameters
    ----------
    target
        observation
    pred
        prediction
    axis
        calculate through which axis

    Returns
    -------
    np.array
        RMSE
    """
    return np.sqrt(np.nanmean((pred - target) ** 2, axis=axis))


def statError(target, pred):
    """
    Calculate multiple statistics indicators:

    Parameters
    ----------
    target
        observation
    pred
        prediction

    Returns
    -------
    dict
        (Bias=Bias, RMSE=RMSE, ubRMSE=ubRMSE, Corr=Corr, R2=R2, NSE=NSE, KGE=KGe, FHV=PBiashigh, FLV=PBiaslow)
    """
    ngrid, nt = pred.shape
    # Bias
    Bias = np.nanmean(pred - target, axis=1)
    # RMSE
    RMSE = np.sqrt(np.nanmean((pred - target) ** 2, axis=1))
    # ubRMSE
    predMean = np.tile(np.nanmean(pred, axis=1), (nt, 1)).transpose()
    targetMean = np.tile(np.nanmean(target, axis=1), (nt, 1)).transpose()
    predAnom = pred - predMean
    targetAnom = target - targetMean
    ubRMSE = np.sqrt(np.nanmean((predAnom - targetAnom) ** 2, axis=1))
    # rho R2 NSE
    Corr = np.full(ngrid, np.nan)
    R2 = np.full(ngrid, np.nan)
    NSE = np.full(ngrid, np.nan)
    KGe = np.full(ngrid, np.nan)
    PBiaslow = np.full(ngrid, np.nan)
    PBiashigh = np.full(ngrid, np.nan)
    PBias = np.full(ngrid, np.nan)
    num_lowtarget_zero = 0
    for k in range(0, ngrid):
        x = pred[k, :]
        y = target[k, :]
        ind = np.where(np.logical_and(~np.isnan(x), ~np.isnan(y)))[0]
        if ind.shape[0] > 0:
            xx = x[ind]
            yy = y[ind]
            # percent bias
            PBias[k] = np.sum(xx - yy) / np.sum(yy) * 100
            if ind.shape[0] > 1:
                # Theoretically at least two points for correlation
                Corr[k] = scipy.stats.pearsonr(xx, yy)[0]
                yymean = yy.mean()
                SST = np.sum((yy - yymean) ** 2)
                SSReg = np.sum((xx - yymean) ** 2)
                SSRes = np.sum((yy - xx) ** 2)
                R2[k] = 1 - SSRes / SST
                NSE[k] = 1 - SSRes / SST
                KGe[k] = KGE(xx, yy)
            # FHV the peak flows bias 2%
            # FLV the low flows bias bottom 30%, log space
            pred_sort = np.sort(xx)
            target_sort = np.sort(yy)
            indexlow = round(0.3 * len(pred_sort))
            indexhigh = round(0.98 * len(pred_sort))
            lowpred = pred_sort[:indexlow]
            highpred = pred_sort[indexhigh:]
            lowtarget = target_sort[:indexlow]
            hightarget = target_sort[indexhigh:]
            if np.sum(lowtarget) == 0:
                num_lowtarget_zero = num_lowtarget_zero + 1
            PBiaslow[k] = np.sum(lowpred - lowtarget) / np.sum(lowtarget) * 100
            PBiashigh[k] = np.sum(highpred - hightarget) / np.sum(hightarget) * 100
            outDict = dict(
                Bias=Bias,
                RMSE=RMSE,
                ubRMSE=ubRMSE,
                Corr=Corr,
                R2=R2,
                NSE=NSE,
                KGE=KGe,
                FHV=PBiashigh,
                FLV=PBiaslow,
            )
    hydro_logger.debug(
        "The CDF of BFLV will not reach 1.0 because some basins have all zero flow observations for the "
        "30% low flow interval, the percent bias can be infinite\n"
        + "The number of these cases is "
        + str(num_lowtarget_zero)
    )
    return outDict



import os

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pytest
import spotpy

import definitions
from hydromodel.calibrate.calibrate_sceua import calibrate_by_sceua, SpotSetup
from hydromodel.models.hymod import hymod
from hydromodel.visual.pyspot_plots import show_calibrate_result


@pytest.fixture()
def basin_area():
    # # the area of basin 01013500 is 2252.7; unit km2
    # the area of a basin from hymod example, unit km2
    return 1.783


@pytest.fixture()
def the_data():
    root_dir = definitions.ROOT_DIR
    return pd.read_csv(
        os.path.join(root_dir, "hydromodel", "example", "hymod_input.csv"), sep=";"
    )
    # return pd.read_csv(os.path.join(root_dir, "hydromodel","example", '01013500_lump_p_pe_q.txt'))


@pytest.fixture()
def qobs(the_data, basin_area):
    # 1 ft3 = 0.02831685 m3
    ft3tom3 = 2.831685e-2
    # 1 km2 = 10^6 m2
    km2tom2 = 1e6
    # 1 m = 1000 mm
    mtomm = 1000
    # 1 day = 24 * 3600 s
    daytos = 24 * 3600
    # qobs_ = np.expand_dims(test_data[['streamflow(ft3/s)']].values, axis=1)
    # trans ft3/s to mm/day
    # return qobs_ * ft3tom3 / (basin_area * km2tom2) * mtomm * daytos

    qobs_ = np.expand_dims(the_data[["Discharge[ls-1]"]].values, axis=1)
    # trans l/s to mm/day
    return qobs_ * 1e-3 / (basin_area * km2tom2) * mtomm * daytos


@pytest.fixture()
def p_and_e(the_data):
    p_and_e_df = the_data[["rainfall[mm]", "TURC [mm d-1]"]]
    # p_and_e_df = test_data[['prcp(mm/day)', 'petfao56(mm/day)']]
    # three dims: batch (basin), sequence (time), feature (variable)
    return np.expand_dims(p_and_e_df.values, axis=1)


@pytest.fixture()
def params():
    return np.array([0.39359, 0.01005, 0.20831, 0.75010, 0.48652]).reshape(1, 5)


def test_hymod(p_and_e, params):
    qsim = hymod(p_and_e, params, warmup_length=10)
    np.testing.assert_almost_equal(
        qsim[:5, 0, 0], [0.0003, 0.0003, 0.0002, 0.0002, 0.0002], decimal=4
    )


def test_calibrate_hymod_sceua(p_and_e, qobs, basin_area):
    calibrate_by_sceua(
        p_and_e,
        qobs,
        model="hymod",
        random_seed=2000,
        rep=5000,
        ngs=7,
        kstop=3,
        peps=0.1,
        pcento=0.1,
    )


def test_show_hymod_calibrate_sceua_result(p_and_e, qobs, basin_area):
    spot_setup = SpotSetup(
        p_and_e,
        qobs,
        warmup_length=10,
        model="hymod",
        obj_func=spotpy.objectivefunctions.rmse,
    )
    show_calibrate_result(spot_setup, "SCEUA_hymod", "l s-1")
    plt.show()



import json
import os
import re
import zipfile
import datetime as dt, datetime
from typing import List
import pickle
from collections import OrderedDict
import numpy as np
import urllib
from urllib import parse

import requests
import matplotlib.pyplot as plt
from itertools import combinations

import threading
import functools

import tqdm

import logging


# -----------------------------------------------logger setting----------------------------------------------------
def get_hydro_logger(log_level_param):
    logger = logging.getLogger(__name__)
    # StreamHandler
    stream_handler = logging.StreamHandler()  # console stream output
    stream_handler.setLevel(level=log_level_param)
    logger.addHandler(stream_handler)
    return logger


log_level = logging.INFO
hydro_logger = get_hydro_logger(log_level)


# ------------------------------------------------progress bar----------------------------------------------------
def provide_progress_bar(
    function, estimated_time, tstep=0.2, tqdm_kwargs={}, args=[], kwargs={}
):
    """
    Tqdm wrapper for a long-running function

    Parameters
    ----------
    function
        function to run
    estimated_time
        how long you expect the function to take
    tstep
        time delta (seconds) for progress bar updates
    tqdm_kwargs
        kwargs to construct the progress bar
    args
        args to pass to the function
    kwargs
        keyword args to pass to the function

    Returns
    -------
    function
    """
    ret = [None]  # Mutable var so the function can store its return value

    def myrunner(function, ret, *args, **kwargs):
        ret[0] = function(*args, **kwargs)

    thread = threading.Thread(
        target=myrunner, args=(function, ret) + tuple(args), kwargs=kwargs
    )
    pbar = tqdm.tqdm(total=estimated_time, **tqdm_kwargs)

    thread.start()
    while thread.is_alive():
        thread.join(timeout=tstep)
        pbar.update(tstep)
    pbar.close()
    return ret[0]


def progress_wrapped(estimated_time, tstep=0.2, tqdm_kwargs={}):
    """Decorate a function to add a progress bar"""

    def real_decorator(function):
        @functools.wraps(function)
        def wrapper(*args, **kwargs):
            return provide_progress_bar(
                function,
                estimated_time=estimated_time,
                tstep=tstep,
                tqdm_kwargs=tqdm_kwargs,
                args=args,
                kwargs=kwargs,
            )

        return wrapper

    return real_decorator


def setup_log(tag="VOC_TOPICS"):
    # create logger
    logger = logging.getLogger(tag)
    # logger.handlers = []
    logger.propagate = False
    logger.setLevel(logging.DEBUG)
    # create console handler and set level to debug
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    # create formatter
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    # add formatter to ch
    ch.setFormatter(formatter)
    # add ch to logger
    # logger.handlers = []
    logger.addHandler(ch)
    return logger


# -------------------------------------------------plot utils-------------------------------------------------------
def save_or_show_plot(file_nm: str, save: bool, save_path=""):
    if save:
        plt.savefig(os.path.join(save_path, file_nm))
    else:
        plt.show()


# ----------------------------------------------------data tools--------------------------------------------------------
def unzip_file(dataset_zip, path_unzip):
    """extract a zip file"""
    with zipfile.ZipFile(dataset_zip, "r") as zip_temp:
        zip_temp.extractall(path_unzip)


def unzip_nested_zip(dataset_zip, path_unzip):
    """Extract a zip file including any nested zip files"""
    with zipfile.ZipFile(dataset_zip, "r") as zfile:
        zfile.extractall(path=path_unzip)
    for root, dirs, files in os.walk(path_unzip):
        for filename in files:
            if re.search(r"\.zip$", filename):
                file_spec = os.path.join(root, filename)
                new_dir = os.path.join(root, filename[0:-4])
                unzip_nested_zip(file_spec, new_dir)


def zip_file_name_from_url(data_url, data_dir):
    data_url_str = data_url.split("/")
    filename = parse.unquote(data_url_str[-1])
    zipfile_path = os.path.join(data_dir, filename)
    unzip_dir = os.path.join(data_dir, filename[0:-4])
    return zipfile_path, unzip_dir


def is_there_file(zipfile_path, unzip_dir):
    """if a file has existed"""
    if os.path.isfile(zipfile_path):
        if os.path.isdir(unzip_dir):
            return True
        unzip_nested_zip(zipfile_path, unzip_dir)
        return True


def download_one_zip(data_url, data_dir):
    """download one zip file from url as data_file"""
    zipfile_path, unzip_dir = zip_file_name_from_url(data_url, data_dir)
    if not is_there_file(zipfile_path, unzip_dir):
        if not os.path.isdir(unzip_dir):
            os.mkdir(unzip_dir)
        r = requests.get(data_url, stream=True)
        with open(zipfile_path, "wb") as py_file:
            for chunk in r.iter_content(chunk_size=1024):  # 1024 bytes
                if chunk:
                    py_file.write(chunk)
        unzip_nested_zip(zipfile_path, unzip_dir), download_small_file


def download_small_zip(data_url, data_dir):
    """download zip file and unzip"""
    zipfile_path, unzip_dir = zip_file_name_from_url(data_url, data_dir)
    if not is_there_file(zipfile_path, unzip_dir):
        if not os.path.isdir(unzip_dir):
            os.mkdir(unzip_dir)
        zipfile_path, _ = urllib.request.urlretrieve(data_url, zipfile_path)
        unzip_nested_zip(zipfile_path, unzip_dir)


def download_small_file(data_url, temp_file):
    """download data from url to the temp_file"""
    r = requests.get(data_url)
    with open(temp_file, "w") as f:
        f.write(r.text)


def download_excel(data_url, temp_file):
    """download a excel file according to url"""
    if not os.path.isfile(temp_file):
        urllib.request.urlretrieve(data_url, temp_file)


def download_a_file_from_google_drive(drive, dir_id, download_dir):
    file_list = drive.ListFile(
        {"q": "'" + dir_id + "' in parents and trashed=false"}
    ).GetList()
    for file in file_list:
        print("title: %s, id: %s" % (file["title"], file["id"]))
        file_dl = drive.CreateFile({"id": file["id"]})
        print("mimetype is %s" % file_dl["mimeType"])
        if file_dl["mimeType"] == "application/vnd.google-apps.folder":
            download_dir_sub = os.path.join(download_dir, file_dl["title"])
            if not os.path.isdir(download_dir_sub):
                os.makedirs(download_dir_sub)
            download_a_file_from_google_drive(drive, file_dl["id"], download_dir_sub)
        else:
            # download
            temp_file = os.path.join(download_dir, file_dl["title"])
            if os.path.isfile(temp_file):
                print("file has been downloaded")
                continue
            file_dl.GetContentFile(os.path.join(download_dir, file_dl["title"]))
            print("Downloading file finished")


def serialize_json(my_dict, my_file):
    with open(my_file, "w") as FP:
        json.dump(my_dict, FP, indent=4)


def unserialize_json_ordered(my_file):
    # m_file = os.path.join(my_file, 'master.json')
    with open(my_file, "r") as fp:
        m_dict = json.load(fp, object_pairs_hook=OrderedDict)
    return m_dict


def unserialize_json(my_file):
    with open(my_file, "r") as fp:
        my_object = json.load(fp)
    return my_object


def serialize_pickle(my_object, my_file):
    f = open(my_file, "wb")
    pickle.dump(my_object, f)
    f.close()


def unserialize_pickle(my_file):
    f = open(my_file, "rb")
    my_object = pickle.load(f)
    f.close()
    return my_object


def serialize_numpy(my_array, my_file):
    np.save(my_file, my_array)


def unserialize_numpy(my_file):
    y = np.load(my_file)
    return y


# -------------------------------------------------time & date tools--------------------------------------------------
def t2dt(t, hr=False):
    t_out = None
    if type(t) is int:
        if 30000000 > t > 10000000:
            t = dt.datetime.strptime(str(t), "%Y%m%d").date()
            t_out = t if hr is False else t.datetime()

    if type(t) is dt.date:
        t_out = t if hr is False else t.datetime()

    if type(t) is dt.datetime:
        t_out = t.date() if hr is False else t

    if t_out is None:
        raise Exception("hydroDL.utils.t2dt failed")
    return t_out


def t_range2_array(t_range, *, step=np.timedelta64(1, "D")):
    sd = t2dt(t_range[0])
    ed = t2dt(t_range[1])
    tArray = np.arange(sd, ed, step)
    return tArray


def t_range_days(t_range, *, step=np.timedelta64(1, "D")):
    sd = dt.datetime.strptime(t_range[0], "%Y-%m-%d")
    ed = dt.datetime.strptime(t_range[1], "%Y-%m-%d")
    t_array = np.arange(sd, ed, step)
    return t_array


def t_days_lst2range(t_array):
    if type(t_array[0]) == np.datetime64:
        t0 = t_array[0].astype(datetime.datetime)
        t1 = t_array[-1].astype(datetime.datetime)
    else:
        t0 = t_array[0]
        t1 = t_array[-1]
    sd = t0.strftime("%Y-%m-%d")
    ed = t1.strftime("%Y-%m-%d")
    return [sd, ed]


def t_range_years(t_range):
    """t_range is a left-closed and right-open interval, if t_range[1] is not Jan.1 then end_year should be included"""
    start_year = int(t_range[0].split("-")[0])
    end_year = int(t_range[1].split("-")[0])
    end_month = int(t_range[1].split("-")[1])
    end_day = int(t_range[1].split("-")[2])
    if end_month == 1 and end_day == 1:
        year_range_list = np.arange(start_year, end_year)
    else:
        year_range_list = np.arange(start_year, end_year + 1)
    return year_range_list


def get_year(a_time):
    if isinstance(a_time, datetime.date):
        return a_time.year
    elif isinstance(a_time, np.datetime64):
        return a_time.astype("datetime64[Y]").astype(int) + 1970
    else:
        return int(a_time[0:4])


def intersect(t_lst1, t_lst2):
    C, ind1, ind2 = np.intersect1d(t_lst1, t_lst2, return_indices=True)
    return ind1, ind2


def date_to_julian(a_time):
    if type(a_time) == str:
        fmt = "%Y-%m-%d"
        dt = datetime.datetime.strptime(a_time, fmt)
    else:
        dt = a_time
    tt = dt.timetuple()
    julian_date = tt.tm_yday
    return julian_date


def t_range_to_julian(t_range):
    t_array = t_range_days(t_range)
    t_array_str = np.datetime_as_string(t_array)
    julian_dates = [date_to_julian(a_time[0:10]) for a_time in t_array_str]
    return julian_dates


# --------------------------------------------------MATH CALCULATION---------------------------------------------------
def subset_of_dict(dict, chosen_keys):
    """make a new dict from key-values of chosen keys in a list"""
    return {key: value for key, value in dict.items() if key in chosen_keys}


def pair_comb(combine_attrs):
    if len(combine_attrs) == 1:
        values = list(combine_attrs[0].values())[0]
        key = list(combine_attrs[0].keys())[0]
        results = []
        for value in values:
            d = dict()
            d[key] = value
            results.append(d)
        return results
    items_all = list()
    for dict_item in combine_attrs:
        list_temp = list(dict_item.values())[0]
        items_all = items_all + list_temp
    all_combs = list(combinations(items_all, 2))

    def is_in_same_item(item1, item2):
        for dict_item in combine_attrs:
            list_now = list(dict_item.values())[0]
            if item1 in list_now and item2 in list_now:
                return True
        return False

    def which_dict(item):
        for dict_item in combine_attrs:
            list_now = list(dict_item.values())[0]
            if item in list_now:
                return list(dict_item.keys())[0]

    combs = [comb for comb in all_combs if not is_in_same_item(comb[0], comb[1])]
    combs_dict = [
        {which_dict(comb[0]): comb[0], which_dict(comb[1]): comb[1]} for comb in combs
    ]
    return combs_dict


def flat_data(x):
    xArrayTemp = x.flatten()
    xArray = xArrayTemp[~np.isnan(xArrayTemp)]
    xSort = np.sort(xArray)
    return xSort


def interpNan(x, mode="linear"):
    if len(x.shape) == 1:
        ngrid = 1
        nt = x.shape[0]
    else:
        ngrid, nt = x.shape
    for k in range(ngrid):
        xx = x[k, :]
        xx = interpNan1d(xx, mode)
    return x


def interpNan1d(x, mode="linear"):
    i0 = np.where(np.isnan(x))[0]
    i1 = np.where(~np.isnan(x))[0]
    if len(i1) > 0:
        if mode == "linear":
            x[i0] = np.interp(i0, i1, x[i1])
        if mode == "pre":
            x0 = x[i1[0]]
            for k in range(len(x)):
                if k in i0:
                    x[k] = x0
                else:
                    x0 = x[k]

    return x


def is_any_elem_in_a_lst(lst1, lst2, return_index=False, include=False):
    do_exist = False
    idx_lst = []
    for j in range(len(lst1)):
        if include:
            for lst2_elem in lst2:
                if lst1[j] in lst2_elem:
                    idx_lst.append(j)
                    do_exist = True
        else:
            if lst1[j] in lst2:
                idx_lst.append(j)
                do_exist = True
    if return_index:
        return do_exist, idx_lst
    return do_exist


def random_choice_no_return(arr, num_lst):
    """sampling without replacement multi-times, and the num of each time is in num_lst"""
    num_lst_arr = np.array(num_lst)
    num_sum = num_lst_arr.sum()
    if type(arr) == list:
        arr = np.array(arr)
    assert num_sum <= arr.size
    results = []
    arr_residue = np.arange(arr.size)
    for num in num_lst_arr:
        idx_chosen = np.random.choice(arr_residue.size, num, replace=False)
        chosen_idx_in_arr = np.sort(arr_residue[idx_chosen])
        results.append(arr[chosen_idx_in_arr])
        arr_residue = np.delete(arr_residue, idx_chosen)
    return results


def find_integer_factors_close_to_square_root(integer):
    start = int(np.sqrt(integer))
    factor = integer / start
    while not is_integer(factor):
        start += 1
        factor = integer / start
    return int(factor), start


def is_integer(number):
    if int(number) == number:
        return True
    else:
        return False


def random_index(ngrid, nt, dim_subset):
    batch_size, rho = dim_subset
    i_grid = np.random.randint(0, ngrid, [batch_size])
    i_t = np.random.randint(0, nt - rho, [batch_size])
    return i_grid, i_t


def flatten_list_function(input_list: List):
    return [item for sublist in input_list for item in sublist]


def nanlog(x):
    if x != x:
        return np.nan
    else:
        return np.log(x)



from typing import Union

import numpy as np
import spotpy
from spotpy.parameter import Uniform, ParameterSet
from spotpy.objectivefunctions import rmse

from hydromodel.models.gr4j import gr4j
from hydromodel.models.hymod import hymod
from hydromodel.models.xaj import xaj


class SpotSetup(object):
    def __init__(self, p_and_e, qobs, warmup_length=30, model="xaj", obj_func=None):
        """
        Set up for Spotpy

        Parameters
        ----------
        p_and_e
            inputs of model
        qobs
            observation data
        warmup_length
            GR4J model need warmup period
        model
            we support "gr4j", "hymod", and "xaj"
        obj_func
            objective function, typically RMSE
        """
        if model == "xaj":
            self.parameter_names = [
                # Allen, R.G., L. Pereira, D. Raes, and M. Smith, 1998.
                # Crop Evapotranspiration, Food and Agriculture Organization of the United Nations,
                # Rome, Italy. FAO publication 56. ISBN 92-5-104219-5. 290p.
                "K",  # ratio of potential evapotranspiration to reference crop evaporation generally from Allen, 1998
                "B",  # The exponent of the tension water capacity curve
                "IM",  # The ratio of the impervious to the total area of the basin
                "UM",  # Tension water capacity in the upper layer
                "LM",  # Tension water capacity in the lower layer
                "DM",  # Tension water capacity in the deepest layer
                "C",  # The coefficient of deep evapotranspiration
                "SM",  # The areal mean of the free water capacity of surface soil layer
                "EX",  # The exponent of the free water capacity curve
                "KI",  # Outflow coefficients of interflow
                "KG",  # Outflow coefficients of groundwater
                "CS",  # The recession constant of channel system
                "L",  # Lag time
                "CI",  # The recession constant of the lower interflow
                "CG",  # The recession constant of groundwater storage
            ]
        elif model == "xaj_mz":
            # use mizuRoute for xaj's surface routing module
            self.parameter_names = [
                # Allen, R.G., L. Pereira, D. Raes, and M. Smith, 1998.
                # Crop Evapotranspiration, Food and Agriculture Organization of the United Nations,
                # Rome, Italy. FAO publication 56. ISBN 92-5-104219-5. 290p.
                "K",  # ratio of potential evapotranspiration to reference crop evaporation generally from Allen, 1998
                "B",  # The exponent of the tension water capacity curve
                "IM",  # The ratio of the impervious to the total area of the basin
                "UM",  # Tension water capacity in the upper layer
                "LM",  # Tension water capacity in the lower layer
                "DM",  # Tension water capacity in the deepest layer
                "C",  # The coefficient of deep evapotranspiration
                "SM",  # The areal mean of the free water capacity of surface soil layer
                "EX",  # The exponent of the free water capacity curve
                "KI",  # Outflow coefficients of interflow
                "KG",  # Outflow coefficients of groundwater
                "A",  # parameter of mizuRoute
                "THETA",  # parameter of mizuRoute
                "CI",  # The recession constant of the lower interflow
                "CG",  # The recession constant of groundwater storage
            ]
        elif model == "gr4j":
            self.parameter_names = ["x1", "x2", "x3", "x4"]
        elif model == "hymod":
            self.parameter_names = ["cmax", "bexp", "alpha", "ks", "kq"]
        else:
            raise NotImplementedError("We don't provide this model now")
        self.model = model
        self.params = []
        for par_name in self.parameter_names:
            # All parameters' range are [0,1], we will transform them to normal range in the model
            self.params.append(Uniform(par_name, low=0.0, high=1.0))
        # Just a way to keep this example flexible and applicable to various examples
        self.obj_func = obj_func
        # Load Observation data from file
        self.p_and_e = p_and_e
        # chose observation data after warmup period
        self.true_obs = qobs[warmup_length:, :, :]
        self.warmup_length = warmup_length

    def parameters(self):
        return spotpy.parameter.generate(self.params)

    def simulation(self, x: ParameterSet) -> Union[list, np.array]:
        """
        run xaj model

        Parameters
        ----------
        x
            the parameters of xaj. This function only has this one parameter.

        Returns
        -------
        Union[list, np.array]
            simulated result from xaj
        """
        # Here the model is started with one parameter combination
        # TODO: Now ParameterSet only support one list, and we only support one basin's calibration now
        # parameter, 2-dim variable: [basin=1, parameter]
        params = np.array(x).reshape(1, -1)
        if self.model == "xaj":
            sim = xaj(self.p_and_e, params, warmup_length=self.warmup_length)
        elif self.model == "xaj_mz":
            sim = xaj(
                self.p_and_e,
                params,
                warmup_length=self.warmup_length,
                route_method="MZ",
            )
        elif self.model == "gr4j":
            sim = gr4j(self.p_and_e, params, warmup_length=self.warmup_length)
        elif self.model == "hymod":
            sim = hymod(self.p_and_e, params, warmup_length=self.warmup_length)
        else:
            raise NotImplementedError("We don't provide this model now")
        return sim[:, 0, 0]

    def evaluation(self) -> Union[list, np.array]:
        """
        read observation values

        Returns
        -------
        Union[list, np.array]
            observation
        """
        # TODO: we only support one basin's calibration now
        return self.true_obs[:, 0, 0]

    def objectivefunction(
        self,
        simulation: Union[list, np.array],
        evaluation: Union[list, np.array],
        params=None,
    ) -> float:
        """
        A user defined objective function to calculate fitness.

        Parameters
        ----------
        simulation:
            simulation results
        evaluation:
            evaluation results
        params:
            parameters leading to the simulation

        Returns
        -------
        float
            likelihood
        """
        # SPOTPY expects to get one or multiple values back,
        # that define the performance of the model run
        if not self.obj_func:
            # This is used if not overwritten by user
            like = rmse(evaluation, simulation)
        else:
            # Way to ensure flexible spot setup class
            like = self.obj_func(evaluation, simulation)
        return like


def calibrate_by_sceua(p_and_e, qobs, warmup_length=30, model="xaj", **sce_ua_param):
    """
    Function for calibrating hymod

    Parameters
    ----------
    p_and_e
        inputs of model
    qobs
        observation data
    warmup_length
        the length of warmup period
    model
        we support "gr4j", "hymod", and "xaj"
    sce_ua_param
        parameters for sce_ua: random seed=2000, rep=5000, ngs=7, kstop=3, peps=0.1, pcento=0.1 (default values)

    Returns
    -------
    None
    """
    random_seed = sce_ua_param["random_seed"]
    rep = sce_ua_param["rep"]
    ngs = sce_ua_param["ngs"]
    kstop = sce_ua_param["kstop"]
    peps = sce_ua_param["peps"]
    pcento = sce_ua_param["pcento"]
    np.random.seed(random_seed)  # Makes the results reproduceable

    # Initialize the xaj example
    # In this case, we tell the setup which algorithm we want to use, so
    # we can use this exmaple for different algorithms
    spot_setup = SpotSetup(
        p_and_e,
        qobs,
        warmup_length=warmup_length,
        model=model,
        obj_func=spotpy.objectivefunctions.rmse,
    )
    # Select number of maximum allowed repetitions
    sampler = spotpy.algorithms.sceua(
        spot_setup, dbname="SCEUA_" + model, dbformat="csv", random_state=random_seed
    )
    # Start the sampler, one can specify ngs, kstop, peps and pcento id desired
    sampler.sample(rep, ngs=ngs, kstop=kstop, peps=peps, pcento=pcento)
    print("Calibrate Finished!")



"""
Core code for XinAnJiang model
"""
import logging
from typing import Union
from collections import OrderedDict
import numpy as np
from numba import jit
from scipy.special import gamma


@jit
def calculate_evap(lm, c, wu0, wl0, prcp, pet) -> tuple[np.array, np.array, np.array]:
    """
    Three-layers evaporation model from "Watershed Hydrologic Simulation" written by Prof. RenJun Zhao.
    The book is Chinese, and its name is 《流域水文模拟》;
    The three-layers evaporation model is described in Page 76;
    The method is same with that in Page 22-23 in "Hydrologic Forecasting (5-th version)" written by Prof. Weimin Bao.
    This book's Chinese name is 《水文预报》

    Parameters
    ----------
    lm
        average soil moisture storage capacity of lower layer
    c
        coefficient of deep layer
    wu0
        initial soil moisture of upper layer; update in each time step
    wl0
        initial soil moisture of lower layer; update in each time step
    prcp
        basin mean precipitation
    pet
        potential evapotranspiration

    Returns
    -------
    tuple[np.array,np.array,np.array]
        eu/el/ed are evaporation from upper/lower/deeper layer, respectively
    """
    eu = np.where(wu0 + prcp >= pet, pet, wu0 + prcp)
    ed = np.where((wl0 < c * lm) & (wl0 < c * (pet - eu)), c * (pet - eu) - wl0, 0.0)
    el = np.where(
        wu0 + prcp >= pet,
        0.0,
        np.where(
            wl0 >= c * lm,
            (pet - eu) * wl0 / lm,
            np.where(wl0 >= c * (pet - eu), c * (pet - eu), wl0),
        ),
    )
    return eu, el, ed


@jit
def calculate_prcp_runoff(b, im, wm, w0, pe) -> tuple[np.array, np.array]:
    """
    Calculates the amount of runoff generated from rainfall after entering the underlying surface.

    Same in "Watershed Hydrologic Simulation" and "Hydrologic Forecasting (5-th version)"

    Parameters
    ----------
    b
        B; exponent coefficient
    im
        IMP; imperiousness coefficient
    wm
        average soil moisture storage capacity
    w0
        initial soil moisture
    pe
        net precipitation

    Returns
    -------
    tuple[np.array,np.array]
        r -- runoff; r_im -- runoff of impervious part
    """
    wmm = wm * (1.0 + b)
    a = wmm * (1.0 - (1.0 - w0 / wm) ** (1.0 / (1.0 + b)))
    if np.isnan(a).any():
        raise ArithmeticError("Please check if w0>wm or b is a negative value!")
    r_cal = np.where(
        pe > 0.0,
        np.where(
            pe + a < wmm,
            # 1e-5 is a precision which we set to guarantee float's calculation is correct
            pe
            - (wm - w0)
            + wm * (1.0 - np.minimum(a + pe, wmm - 1e-5) / wmm) ** (1.0 + b),
            pe - (wm - w0),
        ),
        np.full(pe.shape, 0.0),
    )
    r = np.maximum(r_cal, 0.0)
    # separate impervious part with the other
    r_im_cal = pe * im
    r_im = np.maximum(r_im_cal, 0.0)
    return r, r_im


def calculate_w_storage(
    um, lm, dm, wu0, wl0, wd0, eu, el, ed, pe, r
) -> tuple[np.array, np.array, np.array]:
    """
    Update the soil moisture values of the three layers.

    According to the equation 2.60 in the book《水文预报》

    Parameters
    ----------
    um
        average soil moisture storage capacity of the upper layer
    lm
        average soil moisture storage capacity of the lower layer
    dm
        average soil moisture storage capacity of the deep layer
    wu0
        initial values of soil moisture in upper layer
    wl0
        initial values of soil moisture in lower layer
    wd0
        initial values of soil moisture in deep layer
    eu
        evaporation of the upper layer; it isn't used in this function
    el
        evaporation of the lower layer
    ed
        evaporation of the deep layer
    pe
        net precipitation; it is able to be negative value in this function
    r
        runoff

    Returns
    -------
    tuple[np.array,np.array,np.array]
        wu,wl,wd -- soil moisture in upper, lower and deep layer
    """
    # pe>0: the upper soil moisture was added firstly, then lower layer, and the final is deep layer
    # pe<=0: no additional water, just remove evapotranspiration,
    # but note the case: e >= p > 0
    # (1) if wu0 + p > e, then e = eu (2) else, wu must be zero
    wu = np.where(
        pe > 0.0,
        np.where(wu0 + pe - r < um, wu0 + pe - r, um),
        np.where(wu0 + pe > 0.0, wu0 + pe, 0.0),
    )
    # calculate wd before wl because it is easier to cal using where statement
    wd = np.where(
        pe > 0.0,
        np.where(wu0 + wl0 + pe - r > um + lm, wu0 + wl0 + wd0 + pe - r - um - lm, wd0),
        wd0 - ed,
    )
    # water balance (equation 2.2 in Page 13, also shown in Page 23)
    # if wu0 + p > e, then e = eu; else p must be used in upper layer,
    # so no matter what the case is, el didn't include p, neither ed
    wl = np.where(pe > 0.0, wu0 + wl0 + wd0 + pe - r - wu - wd, wl0 - el)
    # the water storage should be in reasonable range
    wu_ = np.clip(wu, a_min=0.0, a_max=um)
    wl_ = np.clip(wl, a_min=0.0, a_max=lm)
    wd_ = np.clip(wd, a_min=0.0, a_max=dm)
    return wu_, wl_, wd_


def generation(p_and_e, k, b, im, um, lm, dm, c, wu0=None, wl0=None, wd0=None) -> tuple:
    """
    Single-step runoff generation in XAJ.

    Parameters
    ----------
    p_and_e
        precipitation and potential evapotranspiration
    k
        ratio of potential evapotranspiration to reference crop evaporation
    b
        exponent parameter
    um
        average soil moisture storage capacity of the upper layer
    lm
        average soil moisture storage capacity of the lower layer
    dm
        average soil moisture storage capacity of the deep layer
    im
        impermeability coefficient
    c
        coefficient of deep layer
    wu0
        initial values of soil moisture in upper layer
    wl0
        initial values of soil moisture in lower layer
    wd0
        initial values of soil moisture in deep layer

    Returns
    -------
    tuple[tuple, tuple]
        (r, rim, e, pe), (wu, wl, wd); all variables are np.array
    """
    # make sure physical variables' value ranges are correct
    prcp = np.maximum(p_and_e[:, 0], 0.0)
    # get potential evapotranspiration
    pet = np.maximum(p_and_e[:, 1] * k, 0.0)
    # wm
    wm = um + lm + dm
    if wu0 is None:
        # just an initial value
        wu0 = 0.6 * um
    if wl0 is None:
        wl0 = 0.6 * lm
    if wd0 is None:
        wd0 = 0.6 * dm
    w0_ = wu0 + wl0 + wd0

    # w0 need locate in correct range so that following calculation could be right
    # To make sure float data's calculation is correct, we'd better minus a precision (1e-5)
    w0 = np.minimum(w0_, wm - 1e-5)

    # Calculate the amount of evaporation from storage
    eu, el, ed = calculate_evap(lm, c, wu0, wl0, prcp, pet)
    e = eu + el + ed

    # Calculate the runoff generated by net precipitation
    prcp_difference = prcp - e
    pe = np.maximum(prcp_difference, 0.0)
    r, rim = calculate_prcp_runoff(b, im, wm, w0, pe)
    # Update wu, wl, wd
    wu, wl, wd = calculate_w_storage(
        um, lm, dm, wu0, wl0, wd0, eu, el, ed, prcp_difference, r
    )

    return (r, rim, e, pe), (wu, wl, wd)


def sources(pe, r, sm, ex, ki, kg, s0=None, fr0=None) -> tuple:
    """
    Divide the runoff to different sources

    We use the initial version from the paper of the inventor of the XAJ model -- Prof. Renjun Zhao:
    "Analysis of parameters of the XinAnJiang model". Its Chinese name is <<新安江模型参数的分析>>,
    which could be found by searching in "Baidu Xueshu".
    The module's code can also be found in "Watershed Hydrologic Simulation" (WHS) Page 174.
    It is nearly same with that in "Hydrologic Forecasting" (HF) Page 148-149
    We use the period average runoff as input and the unit period is day so we don't need to difference it as books show

    We also provide code for formula from《水文预报》 the fifth version. Page 40-41 and 150-151;
    the procedures in 《工程水文学》 the third version are different we also provide.
    they are in the "sources5mm" function.

    Parameters
    ------------
    pe
        net precipitation
    r
        runoff from xaj_generation
    sm
        areal mean free water capacity of the surface layer
    ex
        exponent of the free water capacity curve
    ki
        outflow coefficients of the free water storage to interflow relationships
    kg
        outflow coefficients of the free water storage to groundwater relationships
    s0
        free water capacity of last period
    fr0
        runoff area of last period

    Return
    ------------
    tuple[tuple, tuple]
        rs -- surface runoff; ri-- interflow runoff; rg -- groundwater runoff;
        s1 -- final free water capacity;
        all variables are numpy array

    """
    # maximum free water storage capacity in a basin
    ms = sm * (1.0 + ex)
    if fr0 is None:
        fr0 = 0.1
    if s0 is None:
        s0 = 0.5 * sm
    precision = 1e-5
    # For free water storage, because s is related to fr and s0 and fr0 are both values of last period,
    # we have to trans the initial value of s from last period to this one.
    # both WHS（流域水文模拟）'s sample code and HF（水文预报） use s = fr0 * s0 / fr.
    # I think they both think free water reservoir as a cubic tank. Its height is s and area of bottom rectangle is fr
    # but the problem is we will have a cubic tank with varying bottom and height, and fixed boundary (sm is fixed)
    # -> so strange !!! I think maybe 2-sources xaj is more interpretable
    # especially when r=0 then fr0=0, the free water cannot disappear immediately, so we have to use s = s0, fr=fr0
    # fr's formula could be found in Eq. 9 in "Analysis of parameters of the XinAnJiang model",
    # Here our r doesn't include rim, so there is no need to remove rim from r; this is also the method in 《水文预报》（HF）
    fr = np.where(r > 0.0, r / pe, fr0)
    if np.isnan(fr).any():
        raise ArithmeticError("Please check pe's data! there may be 0.0")
    ss = np.minimum(fr0 * s0 / fr, sm - precision)
    au = ms * (1.0 - (1.0 - ss / sm) ** (1.0 / (1.0 + ex)))
    if np.isnan(au).any():
        raise ValueError(
            "Error： NaN values detected. Try set clip function or check your data!!!"
        )

    rs = np.where(
        pe > 0.0,
        np.where(
            pe + au < ms,
            # equation 2-85 in HF
            # set precision to guarantee float data's calculation is correct
            fr
            * (
                pe
                - sm
                + ss
                + sm * ((1 - np.minimum(pe + au, ms - precision) / ms) ** (1 + ex))
            ),
            # equation 2-86 in HF
            fr * (pe + ss - sm),
        ),
        np.full(r.shape, 0.0),
    )
    rs = np.clip(rs, a_min=np.full(rs.shape, 0.0), a_max=r)
    # equation 2-87 in HF, some free water leave, so we update free water storage
    s = ss + (r - rs) / fr
    if np.isnan(s).any():
        raise ArithmeticError("Please check fr's data! there may be 0.0")
    s = np.minimum(s, sm)
    # equation 2-88 in HF, next interflow and ground water will be released from the updated free water storage
    # We use the period average runoff as input and the general unit period is day.
    # Hence, we directly use ki and kg rather than ki_{Δt} in books.
    ri = ki * s * fr
    rg = kg * s * fr
    # equation 2-89 in HF; although it looks different with that in WHS, they are actually same
    # Finally, calculate the final free water storage
    s1 = np.clip(s * (1 - ki - kg), a_min=np.full(s.shape, 0.0), a_max=sm)
    return (rs, ri, rg), (s1, fr)


def sources5mm(
    pe,
    runoff,
    sm,
    ex,
    ki,
    kg,
    s0=None,
    fr0=None,
    time_interval_hours=24,
    book="ShuiWenYuBao",
):
    """
    Divide the runoff to different sources according to books -- 《水文预报》 5th edition and 《工程水文学》 3rd edition

    Parameters
    ----------
    pe
        net precipitation
    runoff
        runoff from xaj_generation
    sm
        areal mean free water capacity of the surface layer
    ex
        exponent of the free water capacity curve
    ki
        outflow coefficients of the free water storage to interflow relationships
    kg
        outflow coefficients of the free water storage to groundwater relationships
    s0
        initial free water capacity
    fr0
        initial area of generation
    time_interval_hours
        由于Ki、Kg、Ci、Cg都是以24小时为时段长定义的，需根据时段长转换
    book
        the methods in 《水文预报》 5th edition and 《工程水文学》 3rd edition are different,
        hence, both are provided, and the default is the former -- "ShuiWenYuBao";
        the other one is "GongChengShuiWenXue"

    Returns
    -------
    tuple[tuple, tuple]
        rs_s -- surface runoff; rss_s-- interflow runoff; rg_s -- groundwater runoff;
        (fr_ds[-1], s_ds[-1]): state variables' final value;
        all variables are numpy array
    """
    # 由于Ki、Kg都是以24小时为时段长定义的，需根据时段长转换
    hours_per_day = 24
    # 非整除情况，时段+1
    residue_temp = hours_per_day % time_interval_hours
    if residue_temp != 0:
        residue_temp = 1
    period_num_1d = int(hours_per_day / time_interval_hours) + residue_temp
    # 当kss+kg>1时，根式为偶数运算时，kss_period会成为复数，这里会报错；另外注意分母可能为0，kss不可取0
    # 对kss+kg的取值进行限制，也是符合物理意义的，地下水出流不能超过自身的蓄水。
    kss_period = (1 - (1 - (ki + kg)) ** (1 / period_num_1d)) / (1 + kg / ki)
    kg_period = kss_period * kg / ki

    # 流域最大点自由水蓄水容量深
    smm = sm * (1 + ex)
    if s0 is None:
        s0 = 0.60 * sm
    if fr0 is None:
        fr0 = 0.02
    fr = np.where(pe > 1e-5, runoff / pe, fr0)
    fr = np.clip(fr, 0.001, 1)

    # 净雨分5mm一段进行计算，因为计算时在FS/FR ~ SMF'关系图上开展，即计算在产流面积上开展，所以用PE做净雨.分段为了差分计算更精确。
    if runoff < 5:
        n = 1
    else:
        residue_temp = runoff % 5
        if residue_temp != 0:
            residue_temp = 1
        n = int(runoff / 5) + residue_temp
    # 整除了就是5mm，不整除就少一些，差分每段小了也挺好
    rn = runoff / n
    pen = pe / n
    kss_d = (1 - (1 - (kss_period + kg_period)) ** (1 / n)) / (
        1 + kg_period / kss_period
    )
    kg_d = kss_d * kg_period / kss_period

    rs = rss = rg = 0

    s_ds = []
    fr_ds = []
    s_ds.append(s0)
    fr_ds.append(fr0)

    for j in range(n):
        # 因为产流面积随着自由水蓄水容量的变化而变化，每5mm净雨对应的产流面积肯定是不同的，因此fr是变化的
        fr0_d = fr_ds[j]
        s0_d = s_ds[j]
        fr_d = 1 - (1 - fr) ** (1 / n)
        s_d = fr0_d * s0_d / fr_d

        if book == "ShuiWenYuBao":
            ms = smm
            if s_d > sm:
                s_d = sm
            au = ms * (1 - (1 - s_d / sm) ** (1 / (1 + ex)))
            if pen + au >= ms:
                rs_j = (pen + s_d - sm) * fr_d
            else:
                rs_j = (pen - sm + s_d + sm * (1 - (pen + au) / ms) ** (ex + 1)) * fr_d
            s_d = s_d + (rn - rs_j) / fr_d
            rss_j = s_d * kss_d * fr_d
            rg_j = s_d * kg_d * fr_d
            s_d = s_d * (1 - rss_j + rg_j)

        elif book == "GongChengShuiWenXue":
            smmf = smm * (1 - (1 - fr_d) ** (1 / ex))
            smf = smmf / (1 + ex)
            # 如果出现s_d>smf的情况，说明s_d = fr0_d * s0_d / fr_d导致的计算误差不合理，需要进行修正。
            if s_d > smf:
                s_d = smf
            au = smmf * (1 - (1 - s_d / smf) ** (1 / (1 + ex)))
            if pen + au >= smmf:
                rs_j = (pen + s_d - smf) * fr_d
                rss_j = smf * kss_d * fr_d
                rg_j = smf * kg_d * fr_d
                s_d = smf - (rss_j + rg_j) / fr_d
            else:
                rs_j = (
                    pen - smf + s_d + smf * (1 - (pen + au) / smmf) ** (ex + 1)
                ) * fr_d
                rss_j = (pen - rs_j / fr_d + s_d) * kss_d * fr_d
                rg_j = (pen - rs_j / fr_d + s_d) * kg_d * fr_d
                s_d = s_d + pen - (rs_j + rss_j + rg_j) / fr_d
        else:
            raise NotImplementedError(
                "We don't have this implementation! Please chose 'ShuiWenYuBao' or 'GongChengShuiWenXue'!!"
            )
        rs = rs + rs_j
        rss = rss + rss_j
        rg = rg + rg_j
        # 赋值s_d和fr_d到数组中，以给下一段做初值
        s_ds.append(s_d)
        fr_ds.append(fr_d)

    return (rs, rss, rg), (s_ds[-1], fr_ds[-1])


@jit
def linear_reservoir(x, weight, last_y=None) -> np.array:
    """
    Linear reservoir's release function

    Parameters
    ----------
    x
        the input to the linear reservoir
    weight
        the coefficient of linear reservoir
    last_y
        the output of last period

    Returns
    -------
    np.array
        one-step forward result
    """
    weight1 = 1 - weight
    if last_y is None:
        last_y = np.full(weight.shape, 0.001)
    y = weight * last_y + weight1 * x
    return y


def uh_conv(x, uh_from_gamma):
    """
    Function for 1d-convolution calculation

    Parameters
    ----------
    x
        x is a sequence-first variable; the dim of x is [seq, batch, feature=1];
        feature must be 1
    uh_from_gamma
        unit hydrograph from uh_gamma; the dim: [len_uh, batch, feature=1];
        feature must be 1

    Returns
    -------
    np.array
        convolution
    """
    outputs = np.full(x.shape, 0.0)
    time_length, batch_size, feature_size = x.shape
    if feature_size > 1:
        logging.error("We only support one-dim convolution now!!!")
    for i in range(batch_size):
        uh = uh_from_gamma[:, i, 0]
        inputs = x[:, i, 0]
        outputs[:, i, 0] = np.convolve(inputs, uh)[:time_length]
    return outputs


def uh_gamma(a, theta, len_uh=15):
    """
    A simple two-parameter Gamma distribution as a unit-hydrograph to route instantaneous runoff from a hydrologic model
    The method comes from mizuRoute -- http://www.geosci-model-dev.net/9/2223/2016/

    Parameters
    ----------
    a
        shape parameter
    theta
        timescale parameter
    len_uh
        the time length of the unit hydrograph
    Returns
    -------
    torch.Tensor
        the unit hydrograph, dim: [seq, batch, feature]
    """
    # dims of a: time_seq (same all time steps), batch, feature=1
    m = a.shape
    if len_uh > m[0]:
        raise RuntimeError(
            "length of unit hydrograph should be smaller than the whole length of input"
        )
    # aa > 0, here we set minimum 0.1 (min of a is 0, set when calling this func); First dimension of a is repeat
    aa = np.maximum(0.0, a[0:len_uh, :, :]) + 0.1
    # theta > 0, here set minimum 0.5
    theta = np.maximum(0.0, theta[0:len_uh, :, :]) + 0.5
    # len_f, batch, feature
    t = np.expand_dims(
        np.swapaxes(np.tile(np.arange(0.5, len_uh * 1.0), (m[1], 1)), 0, 1), axis=-1
    )
    denominator = gamma(aa) * (theta**aa)
    # [len_f, m[1], m[2]]
    w = 1 / denominator * (t ** (aa - 1)) * (np.exp(-t / theta))
    w = w / w.sum(0)  # scale to 1 for each UH
    return w


def xaj(
    p_and_e,
    params: Union[np.array, list],
    return_state=False,
    kernel_size=15,
    warmup_length=30,
    route_method="CSL",
    source_type="sources",
    source_book="ShuiWenYuBao",
) -> Union[tuple, np.array]:
    """
    run XAJ model

    Parameters
    ----------
    p_and_e
        prcp and pet; sequence-first (time is the first dim) 3-d np array: [time, basin, feature=2]
    params
        parameters of XAJ model for basin(s);
        2-dim variable -- [basin, parameter]:
        the parameters are B IM UM LM DM C SM EX KI KG A THETA CI CG (notice the sequence)
    return_state
        if True, return state values, mainly for warmup periods
    kernel_size
        the length of unit hydrograph
    warmup_length
        hydro models need a warm-up period to get good initial state values
    route_method
        now we provide two ways: "CSL" (recession constant + lag time) and "MZ" (method from mizuRoute)
    source_type
        default is "sources" and it will call "sources" function; the other is "sources5mm",
        and we will divide the runoff to some <5mm pieces according to the books in this case
    source_book
        When source_type is "sources5mm" there are two implementions for dividing sources,
        as the methods in "ShuiWenYuBao" and "GongChengShuiWenXue"" are different.
        Hence, both are provided, and the default is the former.

    Returns
    -------
    Union[np.array, tuple]
        streamflow or (streamflow, states)
    """
    # params
    if route_method == "CSL":
        param_ranges = OrderedDict(
            {
                "K": [0.5, 2.0],
                "B": [0.1, 0.4],
                "IM": [0.01, 0.1],
                "UM": [0.0, 20.0],
                "LM": [60.0, 90.0],
                "DM": [60.0, 120.0],
                "C": [0.0, 0.2],
                "SM": [1, 100.0],
                "EX": [1.0, 1.5],
                "KI": [0.0, 0.7],
                "KG": [0.0, 0.7],
                "CS": [0.0, 1.0],
                "L": [1.0, 10.0],  # unit is day
                "CI": [0.0, 0.9],
                "CG": [0.98, 0.998],
            }
        )
    elif route_method == "MZ":
        param_ranges = OrderedDict(
            {
                "K": [0.5, 2.0],
                "B": [0.1, 0.4],
                "IM": [0.01, 0.1],
                "UM": [0.0, 20.0],
                "LM": [60.0, 90.0],
                "DM": [60.0, 120.0],
                "C": [0.0, 0.2],
                "SM": [1, 100.0],
                "EX": [1.0, 1.5],
                "KI": [0.0, 0.7],
                "KG": [0.0, 0.7],
                "A": [0.0, 2.9],
                "THETA": [0.0, 6.5],
                "CI": [0.0, 0.9],
                "CG": [0.98, 0.998],
            }
        )
    else:
        raise NotImplementedError(
            "We don't provide this route method now! Please use 'CS' or 'MZ'!"
        )
    xaj_params = [
        (value[1] - value[0]) * params[:, i] + value[0]
        for i, (key, value) in enumerate(param_ranges.items())
    ]
    k = xaj_params[0]
    b = xaj_params[1]
    im = xaj_params[2]
    um = xaj_params[3]
    lm = xaj_params[4]
    dm = xaj_params[5]
    c = xaj_params[6]
    sm = xaj_params[7]
    ex = xaj_params[8]
    ki = xaj_params[9]
    kg = xaj_params[10]
    # ki+kg should be smaller than 1; if not, we scale them
    ki = np.where(ki + kg < 1.0, ki, 1 / (ki + kg) * ki)
    kg = np.where(ki + kg < 1.0, kg, 1 / (ki + kg) * kg)
    if route_method == "CSL":
        cs = xaj_params[11]
        l = xaj_params[12]
    elif route_method == "MZ":
        # we will use routing method from mizuRoute -- http://www.geosci-model-dev.net/9/2223/2016/
        a = xaj_params[11]
        theta = xaj_params[12]
    else:
        raise NotImplementedError(
            "We don't provide this route method now! Please use 'CS' or 'MZ'!"
        )
    ci = xaj_params[13]
    cg = xaj_params[14]

    # initialize state values
    if warmup_length > 0:
        p_and_e_warmup = p_and_e[0:warmup_length, :, :]
        _, *w0, s0, fr0, qi0, qg0 = xaj(
            p_and_e_warmup,
            params,
            return_state=True,
            kernel_size=kernel_size,
            warmup_length=0,
        )
    else:
        w0 = (0.5 * um, 0.5 * lm, 0.5 * dm)
        s0 = 0.5 * sm
        fr0 = np.full(ex.shape, 0.1)
        qi0 = np.full(ci.shape, 0.1)
        qg0 = np.full(cg.shape, 0.1)

    # state_variables
    inputs = p_and_e[warmup_length:, :, :]
    runoff_ims_ = np.full(inputs.shape[:2], 0.0)
    rss_ = np.full(inputs.shape[:2], 0.0)
    ris_ = np.full(inputs.shape[:2], 0.0)
    rgs_ = np.full(inputs.shape[:2], 0.0)
    for i in range(inputs.shape[0]):
        if i == 0:
            (r, rim, e, pe), w = generation(
                inputs[i, :, :], k, b, im, um, lm, dm, c, *w0
            )
            if source_type == "sources":
                (rs, ri, rg), (s, fr) = sources(pe, r, sm, ex, ki, kg, s0, fr0)
            elif source_type == "sources5mm":
                (rs, ri, rg), (s, fr) = sources5mm(
                    pe, r, sm, ex, ki, kg, s0, fr0, book=source_book
                )
            else:
                raise NotImplementedError("No such divide-sources method")
        else:
            (r, rim, e, pe), w = generation(
                inputs[i, :, :], k, b, im, um, lm, dm, c, *w
            )
            if source_type == "sources":
                (rs, ri, rg), (s, fr) = sources(pe, r, sm, ex, ki, kg, s, fr)
            elif source_type == "sources5mm":
                (rs, ri, rg), (s, fr) = sources5mm(
                    pe, r, sm, ex, ki, kg, s, fr, book=source_book
                )
            else:
                raise NotImplementedError("No such divide-sources method")
        # impevious part is pe * im
        runoff_ims_[i, :] = rim
        # so for non-imprvious part, the result should be corrected
        rss_[i, :] = rs * (1 - im)
        ris_[i, :] = ri * (1 - im)
        rgs_[i, :] = rg * (1 - im)
    # seq, batch, feature
    runoff_im = np.expand_dims(runoff_ims_, axis=2)
    rss = np.expand_dims(rss_, axis=2)

    qs = np.full(inputs.shape[:2], 0.0)
    if route_method == "CSL":
        qt = np.full(inputs.shape[:2], 0.0)
        for i in range(inputs.shape[0]):
            if i == 0:
                qi = linear_reservoir(ris_[i], ci, qi0)
                qg = linear_reservoir(rgs_[i], cg, qg0)
            else:
                qi = linear_reservoir(ris_[i], ci, qi)
                qg = linear_reservoir(rgs_[i], cg, qg)
            qs_ = rss_[i]
            qt[i, :] = qs_ + qi + qg
        for j in range(len(l)):
            lag = int(l[j])
            for i in range(lag):
                qs[i, j] = qt[i, j]
            for i in range(lag, inputs.shape[0]):
                qs[i, j] = cs[j] * qs[i - 1, j] + (1 - cs[j]) * qt[i - lag, j]
    elif route_method == "MZ":
        rout_a = a.repeat(rss.shape[0]).reshape(rss.shape)
        rout_b = theta.repeat(rss.shape[0]).reshape(rss.shape)
        conv_uh = uh_gamma(rout_a, rout_b, kernel_size)
        qs_ = uh_conv(runoff_im + rss, conv_uh)
        for i in range(inputs.shape[0]):
            if i == 0:
                qi = linear_reservoir(ris_[i], ci, qi0)
                qg = linear_reservoir(rgs_[i], cg, qg0)
            else:
                qi = linear_reservoir(ris_[i], ci, qi)
                qg = linear_reservoir(rgs_[i], cg, qg)
            qs[i, :] = qs_[i, :, 0] + qi + qg
    else:
        raise NotImplementedError(
            "We don't provide this route method now! Please use 'CS' or 'MZ'!"
        )

    # seq, batch, feature
    q_sim = np.expand_dims(qs, axis=2)
    if return_state:
        return q_sim, *w, s, fr, qi, qg
    return q_sim



"""Calibrate XAJ model using DEAP"""
from deap import base, creator
import random
from deap import tools
import numpy as np
from hydromodel.calibrate.stat import statRmse
from hydromodel.models.gr4j import gr4j
from hydromodel.models.hymod import hymod
from hydromodel.models.xaj import xaj


def evaluate(individual, x_input, y_true, warmup_length, model):
    """
    Calculate fitness for optimization

    Parameters
    ----------
    individual
        individual is the params of XAJ (see details in xaj.py); we initialize all parameters in range [0,1]
    x_input
        input of XAJ
    y_true
        observation data; we use the part after warmup period
    warmup_length
        the length of warmup period
    model
        model's name: "xaj", "xaj_mz", "gr4j", or "hymod"

    Returns
    -------
    float
        fitness
    """
    print("Calculate fitness:")
    # TODO: Now spotpy only support one list, and we only support one basin's calibration now
    params = np.array(individual).reshape(1, -1)
    if model == "xaj":
        sim = xaj(x_input, params, warmup_length=warmup_length)
    elif model == "xaj_mz":
        sim = xaj(x_input, params, warmup_length=warmup_length, route_method="MZ")
    elif model == "gr4j":
        sim = gr4j(x_input, params, warmup_length=warmup_length)
    elif model == "hymod":
        sim = hymod(x_input, params, warmup_length=warmup_length)
    else:
        raise NotImplementedError("We don't provide this model now")
    rmses = statRmse(y_true[warmup_length:, :, :], sim)
    rmse = rmses.mean(axis=0)
    print("-----------------RMSE：" + str(rmse) + "------------------------")
    return rmse


def checkBounds(min, max):
    """
    A decorator to set bounds for individuals in a population

    Parameters
    ----------
    min
        the lower bound of individuals
    max
        the upper bound of individuals

    Returns
    -------
    Function
        a wrapper for clipping data into a given bound
    """

    def decorator(func):
        def wrapper(*args, **kargs):
            offspring = func(*args, **kargs)
            for child in offspring:
                for i in range(len(child)):
                    if child[i] > max:
                        child[i] = max
                    elif child[i] < min:
                        child[i] = min
            return offspring

        return wrapper

    return decorator


MIN = 0
MAX = 1


def calibrate_by_ga(
    input_data, observed_output, warmup_length=30, model="xaj", param_num=15, **ga_param
):
    """
    Use GA algorithm to find optimal parameters for hydrologic models

    Parameters
    ----------
    model
        model's name: "xaj", "xaj_mz", "gr4j", or "hymod"
    input_data
        the input data for model
    observed_output
        the "true" values, i.e. observations
    warmup_length
        the length of warmup period
    param_num
        the number of parameters for model
    ga_param
        run_counts: int = 40, running counts
        pop_num: int = 50, the number of individuals in the population
        cross_prob: float = 0.5, the probability with which two individuals are crossed
        mut_prob: float=0.5, the probability for mutating an individual

    Returns
    -------
    toolbox.population
        optimal_params
    """
    creator.create("FitnessMin", base.Fitness, weights=(-1.0,))
    creator.create("Individual", list, fitness=creator.FitnessMin)
    toolbox = base.Toolbox()
    toolbox.register("attribute", random.random)
    toolbox.register(
        "individual",
        tools.initRepeat,
        creator.Individual,
        toolbox.attribute,
        n=param_num,
    )
    toolbox.register("population", tools.initRepeat, list, toolbox.individual)

    toolbox.register("mate", tools.cxTwoPoint)
    toolbox.register("mutate", tools.mutGaussian, mu=0, sigma=1, indpb=0.1)
    toolbox.register("select", tools.selTournament, tournsize=3)
    toolbox.register(
        "evaluate",
        evaluate,
        x_input=input_data,
        y_true=observed_output,
        warmup_length=warmup_length,
        model=model,
    )

    toolbox.decorate("mate", checkBounds(MIN, MAX))
    toolbox.decorate("mutate", checkBounds(MIN, MAX))

    pop = toolbox.population(n=ga_param["pop_num"])
    # cxpb  is the probability with which two individuals are crossed
    # mutpb is the probability for mutating an individual
    cxpb, mutpb = ga_param["cross_prob"], ga_param["mut_prob"]

    # Evaluate the entire population
    fitnesses = map(toolbox.evaluate, pop)
    for ind, fit in zip(pop, fitnesses):
        ind.fitness.values = fit

    for g in range(ga_param["run_counts"]):
        # Select the next generation individuals
        offspring = toolbox.select(pop, len(pop))
        # Clone the selected individuals
        offspring = list(map(toolbox.clone, offspring))

        # Apply crossover and mutation on the offspring
        for child1, child2 in zip(offspring[::2], offspring[1::2]):
            if random.random() < cxpb:
                toolbox.mate(child1, child2)
                del child1.fitness.values
                del child2.fitness.values

        for mutant in offspring:
            if random.random() < mutpb:
                toolbox.mutate(mutant)
                del mutant.fitness.values

        # Evaluate the individuals with an invalid fitness
        invalid_ind = [ind for ind in offspring if not ind.fitness.valid]
        fitnesses = map(toolbox.evaluate, invalid_ind)
        for ind, fit in zip(invalid_ind, fitnesses):
            ind.fitness.values = fit

        # The population is entirely replaced by the offspring
        pop[:] = offspring
    return pop



#!/usr/bin/env python
# -*- coding:utf-8 -*-

from setuptools import setup, find_packages

setup(
    name="testxaj",  # 输入项目名称
    version="",  # 输入版本号
    keywords=[""],  # 输入关键词
    description="",  # 输入概述
    long_description="update",  # 输入描述

    url="",  # 输入项目Github仓库的链接
    author="",  # 输入作者名字
    author_email="",  # 输入作者邮箱
    license="",  # 此为声明文件，一般填写 MIT_license

    packages=find_packages(),
    include_package_data=True,
    platforms="any",
    install_requires=[""],  # 输入项目所用的包
    python_requires='>=3.7',  # Python版本要求
)



import os
from collections import OrderedDict

import numpy as np
import pandas as pd
import pytest

import definitions
from hydromodel.utils import hydro_utils


@pytest.fixture()
def txt_file():
    return os.path.join(
        definitions.ROOT_DIR, "hydromodel", "example", "01013500_lump_p_pe_q.txt"
    )


@pytest.fixture()
def json_file():
    return os.path.join(definitions.ROOT_DIR, "hydromodel", "example", "data_info.json")


@pytest.fixture()
def npy_file():
    return os.path.join(
        definitions.ROOT_DIR, "hydromodel", "example", "basins_lump_p_pe_q.npy"
    )


def test_save_data(txt_file, json_file, npy_file):
    data = pd.read_csv(txt_file)
    print(data.columns)
    # Note: The units are all mm/day! For streamflow, data is divided by basin's area
    variables = ["prcp(mm/day)", "petfao56(mm/day)", "streamflow(mm/day)"]
    data_info = OrderedDict(
        {
            "time": data["date"].values.tolist(),
            "basin": ["01013500"],
            "variable": variables,
        }
    )
    hydro_utils.serialize_json(data_info, json_file)
    # 1 ft3 = 0.02831685 m3
    ft3tom3 = 2.831685e-2
    # 1 km2 = 10^6 m2
    km2tom2 = 1e6
    # 1 m = 1000 mm
    mtomm = 1000
    # 1 day = 24 * 3600 s
    daytos = 24 * 3600
    # trans ft3/s to mm/day
    basin_area = 2252.7
    data[variables[-1]] = (
        data[["streamflow(ft3/s)"]].values
        * ft3tom3
        / (basin_area * km2tom2)
        * mtomm
        * daytos
    )
    df = data[variables]
    hydro_utils.serialize_numpy(np.expand_dims(df.values, axis=1), npy_file)


def test_load_data(txt_file, npy_file):
    data_ = pd.read_csv(txt_file)
    df = data_[["prcp(mm/day)", "petfao56(mm/day)"]]
    data = hydro_utils.unserialize_numpy(npy_file)[:, :, :2]
    np.testing.assert_array_equal(data, np.expand_dims(df.values, axis=1))



# -*- coding: utf-8 -*-
import numpy as np
from numba import jit


def hymod(p_and_e, parameters, warmup_length=30, return_state=False):
    """
    Run Hymod model

    See https://www.proc-iahs.net/368/180/2015/piahs-368-180-2015.pdf for a scientific paper:
    Quan, Z.; Teng, J.; Sun, W.; Cheng, T. & Zhang, J. (2015): Evaluation of the HYMOD model
    for rainfall–runoff simulation using the GLUE method. Remote Sensing and GIS for Hydrology
    and Water Resources, 180 - 185, IAHS Publ. 368. DOI: 10.5194/piahs-368-180-2015.

    Parameters
    ----------
    p_and_e
        precipitation and potential evapotranspiration, 3-dim variable: [time, basin, feature=1]
    parameters
         five parameters: cmax, bexp, alpha, ks, kq
    warmup_length
        the length of warmup period
    return_state
        if True, return x_slow, x_quick, x_loss, else only return streamflow

    Returns
    -------
    Union[list, np.array]
        streamflow, x_slow, x_quick, x_loss or streamflow
    """
    # parameter, 2-dim variable: [parameter=1, basin]
    cmax_scale = [1.0, 500.0]
    bexp_sacle = [0.1, 2.0]
    alpha_scale = [0.1, 0.99]
    ks_scale = [0.001, 0.10]
    kq_scale = [0.1, 0.99]
    cmax = cmax_scale[0] + parameters[:, 0] * (cmax_scale[1] - cmax_scale[0])
    bexp = bexp_sacle[0] + parameters[:, 1] * (bexp_sacle[1] - bexp_sacle[0])
    alpha = alpha_scale[0] + parameters[:, 2] * (alpha_scale[1] - alpha_scale[0])
    ks = ks_scale[0] + parameters[:, 3] * (ks_scale[1] - ks_scale[0])
    kq = kq_scale[0] + parameters[:, 4] * (kq_scale[1] - kq_scale[0])
    if warmup_length > 0:
        # set no_grad for warmup periods
        p_and_e_warmup = p_and_e[0:warmup_length, :, :]
        _, x_slow, x_quick, x_loss = hymod(
            p_and_e_warmup, parameters, warmup_length=0, return_state=True
        )
    else:
        # Initialize slow tank state
        # x_slow = 2.3503 / (ks * 22.5)
        x_slow = np.full(
            (p_and_e.shape[1], 1), 0.0
        )  # --> works ok if calibration data starts with low discharge
        # Initialize state(s) of quick tank(s)
        x_quick = np.full((p_and_e.shape[1], 3), 0.0)
        # HYMOD PROGRAM IS SIMPLE RAINFALL RUNOFF MODEL
        x_loss = np.full((p_and_e.shape[1], 1), 0.0)
    precip = p_and_e[warmup_length:, :, 0]
    pet = p_and_e[warmup_length:, :, 1]
    t = 0
    output = np.full(precip.shape, 0.0)
    # START PROGRAMMING LOOP WITH DETERMINING RAINFALL - RUNOFF AMOUNTS
    while t <= precip.shape[0] - 1:
        pval = precip[t, :]
        pet_val = pet[t, :]
        # Compute excess precipitation and evaporation
        er1, er2, x_loss = excess(x_loss, cmax, bexp, pval, pet_val)
        # Calculate total effective rainfall
        et = er1 + er2
        #  Now partition ER between quick and slow flow reservoirs
        uq = alpha * et
        us = (1 - alpha) * et
        # Route slow flow component with single linear reservoir
        x_slow, qs = linres(x_slow, us, ks)
        # Route quick flow component with linear reservoirs
        inflow = uq

        for i in range(3):
            # Linear reservoir
            x_quick[:, i], outflow = linres(x_quick[:, i], inflow, kq)
            inflow = outflow

        # Compute total flow for timestep
        output[t, :] = qs + outflow
        t = t + 1
    streamflow = np.expand_dims(output, axis=2)
    if return_state:
        return streamflow, x_slow, x_quick, x_loss
    return streamflow


@jit
def power(x, y):
    x = np.abs(x)  # Needed to capture invalid overflow with negative values
    return x**y


@jit
def linres(x_slow, inflow, rs):
    # Linear reservoir
    x_slow = (1 - rs) * x_slow + (1 - rs) * inflow
    outflow = (rs / (1 - rs)) * x_slow
    return x_slow, outflow


@jit
def excess(x_loss, cmax, bexp, pval, pet_val):
    # this function calculates excess precipitation and evaporation
    xn_prev = x_loss
    ct_prev = cmax * (
        1 - power((1 - ((bexp + 1) * (xn_prev) / cmax)), (1 / (bexp + 1)))
    )
    # Calculate Effective rainfall 1
    er1 = np.maximum((pval - cmax + ct_prev), 0.0)
    pval = pval - er1
    dummy = np.minimum(((ct_prev + pval) / cmax), 1)
    xn = (cmax / (bexp + 1)) * (1 - power((1 - dummy), (bexp + 1)))

    # Calculate Effective rainfall 2
    er2 = np.maximum(pval - (xn - xn_prev), 0)

    # Alternative approach
    evap = (
        1 - (((cmax / (bexp + 1)) - xn) / (cmax / (bexp + 1)))
    ) * pet_val  # actual ET is linearly related to the soil moisture state
    xn = np.maximum(xn - evap, 0)  # update state

    return er1, er2, xn



import os

import numpy as np
import pandas as pd
import pytest
import spotpy
from matplotlib import pyplot as plt
import definitions
from hydromodel.calibrate.calibrate_sceua import calibrate_by_sceua, SpotSetup
from hydromodel.models.gr4j import gr4j
from hydromodel.visual.pyspot_plots import show_calibrate_result


@pytest.fixture()
def basin_area():
    # the area of basin 01013500, unit km2
    # basin_area = 2252.7
    return 1.783


@pytest.fixture()
def warmup_length():
    return 30


@pytest.fixture()
def the_data():
    root_dir = definitions.ROOT_DIR
    # test_data = pd.read_csv(os.path.join(root_dir, "hydromodel", "example", '01013500_lump_p_pe_q.txt'))
    return pd.read_csv(
        os.path.join(root_dir, "hydromodel", "example", "hymod_input.csv"), sep=";"
    )


@pytest.fixture()
def p_and_e(the_data):
    # p_and_e_df = test_data[['prcp(mm/day)', 'petfao56(mm/day)']]
    # three dims: sequence (time), batch (basin), feature (variable)
    # p_and_e = np.expand_dims(p_and_e_df.values, axis=1)
    p_and_e_df = the_data[["rainfall[mm]", "TURC [mm d-1]"]]
    return np.expand_dims(p_and_e_df.values, axis=1)


@pytest.fixture()
def qobs(basin_area, the_data):
    # 1 ft3 = 0.02831685 m3
    ft3tom3 = 2.831685e-2
    # 1 km2 = 10^6 m2
    km2tom2 = 1e6
    # 1 m = 1000 mm
    mtomm = 1000
    # 1 day = 24 * 3600 s
    daytos = 24 * 3600
    # qobs_ = np.expand_dims(test_data[['streamflow(ft3/s)']].values, axis=1)
    # trans ft3/s to mm/day
    # return qobs_ * ft3tom3 / (basin_area * km2tom2) * mtomm * daytos

    qobs_ = np.expand_dims(the_data[["Discharge[ls-1]"]].values, axis=1)
    # trans l/s to mm/day
    return qobs_ * 1e-3 / (basin_area * km2tom2) * mtomm * daytos


@pytest.fixture()
def params():
    # all parameters are in range [0,1]
    return np.tile([0.5], (1, 4))


def test_gr4j(p_and_e, params):
    qsim = gr4j(p_and_e, params, warmup_length=10)
    np.testing.assert_array_equal(qsim.shape, (1817, 1, 1))


def test_calibrate_gr4j_sceua(p_and_e, qobs, warmup_length):
    calibrate_by_sceua(
        p_and_e,
        qobs,
        warmup_length,
        model="gr4j",
        random_seed=2000,
        rep=5000,
        ngs=7,
        kstop=3,
        peps=0.1,
        pcento=0.1,
    )


def test_show_calibrate_sceua_result(p_and_e, qobs, warmup_length):
    spot_setup = SpotSetup(
        p_and_e,
        qobs,
        warmup_length,
        model="gr4j",
        obj_func=spotpy.objectivefunctions.rmse,
    )
    show_calibrate_result(spot_setup, "SCEUA_gr4j")
    plt.show()
